"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const dc_configs_1 = require("dc-configs");
const path_1 = __importDefault(require("path"));
const node_fetch_1 = __importDefault(require("node-fetch"));
const dc_core_1 = require("dc-core");
const dc_ethereum_utils_1 = require("dc-ethereum-utils");
const dc_logging_1 = require("dc-logging");
const FileUtils_1 = require("./FileUtils");
const events_1 = require("events");
const PingService_1 = require("./PingService");
const logger = new dc_logging_1.Logger("Bankroller:");
class Bankroller extends events_1.EventEmitter {
    constructor() {
        super();
        const { platformId, blockchainNetwork } = dc_configs_1.config.default;
        this._platformId = platformId;
        // this._platformIdHash = createHash(platformId)
        this._blockchainNetwork = blockchainNetwork;
        this.gamesMap = new Map();
        this.gamesPath = new Map();
        this._loadedDirectories = new Set();
        global.DCLib = new dc_core_1.GlobalGameLogicStore();
    }
    _getApiRoomAddress(ethAddress) {
        return `${this._platformId}_${this._blockchainNetwork}_${ethAddress}`;
    }
    getApiRoomAddress() {
        return this._apiRoomAddress;
    }
    getPlatformId() {
        return this._platformId;
    }
    async start(transportProvider) {
        const { platformId, gasPrice: price, gasLimit: limit, web3HttpProviderUrl: httpProviderUrl, walletName, blockchainNetwork, privateKey, contracts } = dc_configs_1.config.default;
        const ERC20ContractInfo = contracts.ERC20;
        this._eth = new dc_ethereum_utils_1.Eth({
            walletName,
            httpProviderUrl,
            ERC20ContractInfo,
            gasParams: { price, limit }
        });
        if (this._started) {
            throw new Error("Bankroller allready started");
        }
        this._transportProvider = transportProvider;
        await this._eth.initAccount(privateKey);
        await this._eth.saveWallet(privateKey);
        //  check balance
        const balances = await this._eth.getBalances();
        if (balances.bet.balance === 0) {
            throw new Error(`Empty bet balance at address: ${this._eth.getAccount().address}`);
        }
        if (balances.eth.balance < 0.01) {
            throw new Error(`Not enough ETH balance at address: ${this._eth.getAccount().address}`);
        }
        const ethAddress = this._eth.getAccount().address.toLowerCase();
        this._apiRoomAddress = this._getApiRoomAddress(ethAddress);
        transportProvider.exposeSevice(this._apiRoomAddress, this, true);
        this._pingService = new PingService_1.PingService().start(transportProvider, {
            platformId: this._platformId,
            apiRoomAddress: this._apiRoomAddress,
            blockchainNetwork: this._blockchainNetwork,
            ethAddress
        });
        // transportProvider.exposeSevice(this.getPlatformIdHash(), PingService, true)
        this._started = true;
        const subDirectories = FileUtils_1.getSubDirectories(dc_configs_1.config.default.DAppsPath);
        for (let i = 0; i < subDirectories.length; i++) {
            await this.tryLoadDApp(subDirectories[i]);
        }
        logger.info(`Bankroller started. Api address: \x1b[32m${this._apiRoomAddress}\x1b[0m`);
        const stopBankroller = async () => {
            const status = await this.stop();
            process.exit(status ? 0 : 1);
        };
        process.on("SIGTERM", stopBankroller);
        process.on("SIGINT", stopBankroller);
        return this;
    }
    isStarted() {
        return this._started;
    }
    async stop() {
        await this._pingService.stop();
        const status = await this._transportProvider.stopService(this._apiRoomAddress);
        if (status) {
            logger.info(`Bankroller stoped. Api address: ${this._apiRoomAddress}`);
            this._started = false;
        }
        return status;
    }
    async uploadGame({ name, files, reload = false }) {
        const DAppsPath = dc_configs_1.config.default.DAppsPath;
        const newDir = path_1.default.join(DAppsPath, name);
        let status = Bankroller.STATUS_SUCCESS;
        if (reload && this._loadedDirectories.has(newDir)) {
            await this.unloadGame(name);
        }
        if (this._loadedDirectories.has(newDir)) {
            throw new Error(`Directory ${newDir} allready created`);
        }
        if (!FileUtils_1.saveFilesToNewDir(newDir, files)) {
            throw new Error(`Error save files to ${newDir}`);
        }
        this._loadedDirectories.add(newDir);
        if (!(await this.tryLoadDApp(newDir))) {
            FileUtils_1.removeDir(newDir);
            this._loadedDirectories.delete(newDir);
            status = Bankroller.STATUS_FAILURE;
        }
        return { status };
    }
    async unloadGame(name) {
        const DAppsPath = dc_configs_1.config.default.DAppsPath;
        const newDir = path_1.default.join(DAppsPath, name);
        let status = false;
        if (await this.tryUnloadDApp(newDir)) {
            FileUtils_1.removeDir(newDir);
            this._loadedDirectories.delete(newDir);
            status = true;
        }
        return {
            status: status ? Bankroller.STATUS_SUCCESS : Bankroller.STATUS_FAILURE
        };
    }
    getGames() {
        const games = [];
        for (const [slug, dapp] of this.gamesMap) {
            games.push({
                name: slug,
                path: this.gamesPath.get(slug)
            });
        }
        return games;
    }
    getGameInstances(name) {
        const dapp = this.gamesMap.get(name);
        if (!dapp) {
            throw new Error(`Game ${name} not found`);
        }
        return dapp.getInstancesView();
    }
    async tryLoadDApp(directoryPath) {
        const now = Date.now();
        // if (this._loadedDirectories.has(directoryPath)) {
        //   throw new Error(`Directory ${directoryPath} allready loadeed`)
        // }
        try {
            const { gameLogicFunction, manifest } = await FileUtils_1.loadLogic(directoryPath);
            const roomProvider = this._transportProvider;
            if (typeof gameLogicFunction !== "function") {
                throw new Error("gameLogic is not a function");
            }
            const { disabled, slug, rules, contract: manifestContract, getContract } = manifest;
            if (manifest.disabled) {
                logger.debug(`DApp ${slug} disabled - skip`);
                return null;
            }
            let gameContractAddress = manifestContract || getContract(this._blockchainNetwork).address;
            if (gameContractAddress.indexOf("->") > -1 &&
                this._blockchainNetwork === "local") {
                const { web3HttpProviderUrl } = dc_configs_1.config.default;
                gameContractAddress = await node_fetch_1.default(`${web3HttpProviderUrl}/${gameContractAddress.split("->")[0]}`)
                    .then(result => result.json())
                    .then(result => result[gameContractAddress.split("->")[1]]);
            }
            const dapp = new dc_core_1.DApp({
                platformId: this._platformId,
                blockchainNetwork: this._blockchainNetwork,
                slug,
                rules,
                gameLogicFunction,
                gameContractAddress,
                roomProvider,
                Eth: this._eth
            });
            await dapp.startServer();
            this.gamesMap.set(slug, dapp);
            // Это нужно что бы использовать unloadGame удаленно
            const DAppsPath = dc_configs_1.config.default.DAppsPath;
            this.gamesPath.set(slug, directoryPath.replace(DAppsPath, ""));
            logger.debug(`Load Dapp ${directoryPath}, took ${Date.now() - now} ms`);
            // broadcast result to uploadGame
            this.emit(Bankroller.EVENT_UPLOAD_GAME, {
                name: slug,
                path: this.gamesPath.get(slug)
            });
            return dapp;
        }
        catch (error) {
            logger.debug(error);
        }
        return null;
    }
    async tryUnloadDApp(directoryPath) {
        if (!this._loadedDirectories.has(directoryPath)) {
            throw new Error(`Directory ${directoryPath} not loadeed`);
        }
        const now = Date.now();
        try {
            const { gameLogicFunction, manifest } = await FileUtils_1.loadLogic(directoryPath);
            if (gameLogicFunction) {
                const { disabled, slug } = manifest;
                if (!disabled) {
                    // const dapp = this.gamesMap.get(slug)
                    // await dapp.stopServer() // TODO: !!! need code
                    this.emit(Bankroller.EVENT_UNLOAD_GAME, {
                        name: slug,
                        path: this.gamesPath.get(slug)
                    });
                    this.gamesMap.delete(slug);
                    this.gamesPath.delete(slug);
                    logger.debug(`Unload Dapp ${directoryPath}, took ${Date.now() - now} ms`);
                }
                else {
                    logger.debug(`DApp ${slug} disabled - skip`);
                }
                return true;
            }
        }
        catch (error) {
            logger.debug(error);
            return false;
        }
    }
    eventNames() {
        return [Bankroller.EVENT_UNLOAD_GAME, Bankroller.EVENT_UPLOAD_GAME];
    }
}
Bankroller.STATUS_SUCCESS = "ok";
Bankroller.STATUS_FAILURE = "fail";
Bankroller.EVENT_UPLOAD_GAME = "uploadGame";
Bankroller.EVENT_UNLOAD_GAME = "unloadGame";
exports.default = Bankroller;
//# sourceMappingURL=Bankroller.js.map